# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import numpy as np
import tensorflow as tf

from kamae.tensorflow.layers.base import BaseLayer
from kamae.tensorflow.utils import listify_tensors


class NormalizeLayer(BaseLayer):
    """
    Intermediate layer for normalization layers.

    Reduces code duplication by providing a common interface for normalization layers.
    """

    def __init__(
        self,
        mean,
        variance,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        axis: int = -1,
        **kwargs,
    ) -> None:
        """
        Initializes the NormalizeLayer

        :param mean: The mean value(s) to use during normalization. The passed value(s)
        will be broadcast to the shape of the kept axes above; if the value(s)
        cannot be broadcast, an error will be raised when this layer's
        `build()` method is called.
        :param variance: The variance value(s) to use during normalization. The passed
        value(s) will be broadcast to the shape of the kept axes above; if the
        value(s) cannot be broadcast, an error will be raised when this
        layer's `build()` method is called.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param axis: Integer, tuple of integers, or None. The axis or axes that should
        have a separate mean and variance for each index in the shape. For
        example, if shape is `(None, 5)` and `axis=1`, the layer will track 5
        separate mean and variance values for the last axis. If `axis` is set
        to `None`, the layer will normalize all elements in the input by a
        scalar mean and variance. Defaults to -1, where the last axis of the
        input is assumed to be a feature dimension and is normalized per
        index. Note that in the specific case of batched scalar inputs where
        the only axis is the batch axis, the default will normalize each index
        in the batch separately. In this case, consider passing `axis=None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        # Standardize `axis` to a tuple.
        if axis is None:
            axis = ()
        elif isinstance(axis, int):
            axis = (axis,)
        else:
            axis = tuple(axis)

        self.axis = axis
        self.input_mean = mean
        self.input_variance = variance
        self.epsilon = 1e-8

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.bfloat16, tf.float16, tf.float32, tf.float64]

    def build(self, input_shape) -> None:
        """
        Builds shapes for the mean and variance tensors.

        Specifically, understands which axis to compute the normalization across
        and broadcasts the mean and variance tensors to match the input shape.

        :param input_shape: The shape of the input tensor.
        :returns: None - layer is built.
        """
        super().build(input_shape)

        if isinstance(input_shape, (list, tuple)) and all(
            isinstance(shape, (tf.TensorShape, list, tuple)) for shape in input_shape
        ):
            # This seems to be needed to handle sending in multiple inputs as a list.
            # Although this layer should only have one input, so this is a bit of a
            # hack. We catch this nicely in call method with a decorator. Maybe we
            # should do the same here?
            input_shape = input_shape[0]

        input_shape = tf.TensorShape(input_shape).as_list()
        ndim = len(input_shape)
        self._build_input_shape = input_shape

        if any(a < -ndim or a >= ndim for a in self.axis):
            raise ValueError(
                f"""All `axis` values must be in the range [-ndim, ndim). "
                Found ndim: `{ndim}`, axis: {self.axis}"""
            )

        # Axes to be kept, replacing negative values with positive equivalents.
        # Sorted to avoid transposing axes.
        keep_axis = sorted([d if d >= 0 else d + ndim for d in self.axis])
        # All axes to be kept should have known shape.
        for d in keep_axis:
            if input_shape[d] is None:
                raise ValueError(
                    f"""All `axis` values to be kept must have known shape. "
                    Got axis: {self.axis},
                    input shape: {input_shape}, with unknown axis at index: {d}"""
                )
        # Broadcast any reduced axes.
        broadcast_shape = [input_shape[d] if d in keep_axis else 1 for d in range(ndim)]
        mean_and_var_shape = tuple(input_shape[d] for d in keep_axis)
        mean = self.input_mean * np.ones(mean_and_var_shape)
        variance = self.input_variance * np.ones(mean_and_var_shape)
        self.mean = tf.reshape(mean, broadcast_shape)
        self.variance = tf.reshape(variance, broadcast_shape)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the StandardScaleLayer layer.
        Used for saving and loading from a model.
        Specifically adds additional parameters to the base configuration.
        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        # Ensure mean and variance are lists for serialization.
        config.update(
            {
                "mean": listify_tensors(self.input_mean),
                "variance": listify_tensors(self.input_variance),
                "axis": self.axis,
            }
        )
        return config

    def get_build_config(self):
        if self._build_input_shape:
            return {"input_shape": self._build_input_shape}

    def build_from_config(self, config):
        if config:
            self.build(config["input_shape"])



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union

import tensorflow as tf
from tensorflow.keras.layers import StringLookup

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class StringIndexLayer(BaseLayer):
    """
    Wrapper around the Keras StringLookup layer.

    This layer translates a set of arbitrary strings into integer output via a
    table-based vocabulary lookup. This layer will perform no splitting or
    transformation of input strings.
    """

    def __init__(
        self,
        vocabulary: Union[str, List[str]],
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        num_oov_indices: int = 1,
        mask_token: str = None,
        encoding: str = "utf-8",
        **kwargs,
    ) -> None:
        """
        Intialise the StringIndexLayer layer.

        :param vocabulary: Either an array of strings or a string path to a
        text file. If passing an array, can pass a tuple, list, 1D numpy array,
        or 1D tensor containing the string vocbulary terms. If passing a file
        path, the file should contain one line per term in the vocabulary.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param num_oov_indices: The number of out-of-vocabulary tokens to use. If this
        value is more than 1, OOV inputs are hashed to determine their OOV
        value. If this value is 0, OOV inputs will cause an error when calling
        the layer.  Defaults to 1.
        :param mask_token: A token that represents masked inputs. The token is included
        in vocabulary and mapped to index 0. If set to None, no mask term will be added.
        Defaults to `None`.
        :param encoding: Optional. The text encoding to use to interpret the input
        strings. Defaults to `"utf-8"`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.vocabulary = vocabulary
        self.num_oov_indices = num_oov_indices
        self.mask_token = mask_token
        self.encoding = encoding
        self.indexer = StringLookup(
            vocabulary=vocabulary,
            num_oov_indices=num_oov_indices,
            mask_token=mask_token,
            encoding=encoding,
        )

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs string indexing by calling the StringLookup layer.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Input string tensor to index.
        :returns: Indexed tensor.
        """
        return self.indexer(inputs)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the StringIndexer layer.
        Used for saving and loading from a model.

        Specifically adds the `vocabulary`, `num_oov_indices`, `mask_token`, and
        `encoding` to the config.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "vocabulary": self.vocabulary,
                "num_oov_indices": self.num_oov_indices,
                "mask_token": self.mask_token,
                "encoding": self.encoding,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class StringMapLayer(BaseLayer):
    """
    StringMapLayer layer for TensorFlow.
    """

    def __init__(
        self,
        string_match_values: List[str],
        string_replace_values: List[str],
        default_replace_value: str = None,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initialises the StringMapLayer layer.

        :param string_match_values: The list of strings to match against.
        :param string_replace_values: The list of strings to replace the matched
        strings with.
        :param default_replace_value: The default value to replace the unmatched
        strings with. If None, the original string is kept unchanged.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.string_match_values = string_match_values
        self.string_replace_values = string_replace_values
        self.default_replace_value = default_replace_value

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Checks if the input tensor is matching any of the string_match_values
        and replaces it with the corresponding string_replace_values.

        If default_replace_value is set, it will replace the unmatched strings
        with the default_replace_value. If default_replace_value is None, the
        original string is kept unchanged.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param: inputs: Input string tensor.
        :returns: A string tensor with the matched strings replaced.
        """

        # Iterate through each match/replace pair
        output_tensor = inputs
        for match_value, replace_value in zip(
            self.string_match_values, self.string_replace_values
        ):
            output_tensor = tf.where(
                tf.equal(output_tensor, match_value), replace_value, output_tensor
            )

        # Handle the default replacement for unmatched strings
        # Chain tf.logical_and for each match to check if there is no match
        if self.default_replace_value is not None:
            matches = self.string_match_values
            unmatched_condition = tf.not_equal(inputs, matches[0])
            if len(matches) > 1:
                for match in matches[1:]:
                    unmatched_condition = tf.logical_and(
                        unmatched_condition,
                        tf.not_equal(inputs, match),
                    )
            expected_dtype = output_tensor.dtype
            default_val = tf.constant(self.default_replace_value, dtype=expected_dtype)
            output_tensor = tf.where(unmatched_condition, default_val, output_tensor)

        return output_tensor

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the StringMapLayer layer.
        Used for saving and loading the layer from disk.

        Specifically, `string_match_values` and `string_replace_values`
        are added to the config.

        :returns: Dictionary configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "string_match_values": self.string_match_values,
                "string_replace_values": self.string_replace_values,
                "default_replace_value": self.default_replace_value,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

import pyspark.sql.functions as F
from pyspark.sql import Column, DataFrame, Window, WindowSpec
from pyspark.sql.types import ArrayType


def check_listwise_columns(
    dataset: DataFrame,
    query_col_name: str,
    value_col_name: str,
    sort_col_name: Optional[str] = None,
) -> None:
    """
    Check if the query and value columns are array columns in the dataset.
    If so throw an error since these do not support arrays currently.

    :param dataset: DataFrame containing the query and value columns.
    :param query_col_name: Name of the query column.
    :param value_col_name: Name of the value column.

    :raises ValueError: If the query or value column is an array column.
    :returns: None
    """
    query_col_datatype = dataset.schema[query_col_name].dataType
    value_col_datatype = dataset.schema[value_col_name].dataType

    if isinstance(query_col_datatype, ArrayType):
        raise ValueError("Query column cannot be an array column")

    if isinstance(value_col_datatype, ArrayType):
        raise ValueError("Value column cannot be an array column")

    if sort_col_name is not None:
        sort_col_datatype = dataset.schema[sort_col_name].dataType
        if isinstance(sort_col_datatype, ArrayType):
            raise ValueError("Sort column cannot be an array column")


def get_listwise_condition_and_window(
    query_col: Column,
    value_col: Column,
    sort_col: Column = None,
    sort_order: str = "asc",
    sort_top_n: int = None,
    min_filter_value: float = None,
) -> (Column, WindowSpec):
    """
    Get the condition and window operations for listwise statistics calculation.

    :param query_col: Column containing the query id.
    :param value_col: Column containing the value to calculate statistics on.
    :param sort_col: Column to sort the values by. Default is None.
    :param sort_order: Order to sort the values by. Default is "asc".
    :param sort_top_n: Number of top values to consider for statistics calculation.
    Default is None.
    :param min_filter_value: Minimum value to consider for statistics calculation.
    Default is None.
    :returns: Tuple of the condition and window operations.
    """
    condition_col = None
    window_spec = Window.partitionBy(query_col)

    # Define statistics calculation condition based on topN and sortOrder
    if sort_col is not None:
        if sort_order == "asc":
            sort_col = sort_col.asc()
        elif sort_order == "desc":
            sort_col = sort_col.desc()
        else:
            ValueError(f"Invalid sortOrder: {sort_order}")
        rank_fun = F.row_number().over(window_spec.orderBy(sort_col))
        if sort_top_n is not None:
            condition_col = rank_fun <= sort_top_n
        else:
            ValueError("sortTopN must be set if sortCol is set")
    else:
        condition_col = F.lit(True)

    # Define statistics calculation condition based on min filter value
    if min_filter_value is not None:
        condition_col = condition_col & (value_col >= F.lit(min_filter_value))

    return condition_col, window_spec



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class StringContainsListLayer(BaseLayer):
    """
    Performs a string contains operation on the input tensor over entries in
    the string constant list.

    This implementation does not support matching of newline characters or empty
    strings.
    """

    def __init__(
        self,
        string_constant_list: List[str],
        negation: bool = False,
        name: Optional[str] = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initialises the StringContainsListLayer layer.
        :param string_constant_list: The string to match against.
        :param negation: Whether to negate the output. Defaults to `False`.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.negation = negation
        self.string_constant_list = string_constant_list

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Checks for the existence of any substring in the string_contains_list
        within a tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param: inputs: Input string tensor.
        :returns: A boolean tensor indicating whether any of the string constants are
        matched.
        """
        match_substring = "|".join(
            [
                "(.*" + self._escape_special_characters(x) + ".*)"
                for x in self.string_constant_list
            ]
        )
        matched_tensor = tf.strings.regex_full_match(
            inputs,
            match_substring,
        )

        output_tensor = (
            tf.math.logical_not(matched_tensor) if self.negation else matched_tensor
        )

        return output_tensor

    def _escape_special_characters(self, string: str) -> str:
        """
        Escapes special characters in a string so they are not parsed as regex.
        :param string: The string or string tensor to escape special characters in.
        :returns: The escaped string or string tensor.
        """
        escaped_string = string
        for char in [
            "\\",
            ".",
            "^",
            "$",
            "*",
            "+",
            "?",
            "{",
            "}",
            "[",
            "]",
            "(",
            ")",
            "|",
        ]:
            if isinstance(escaped_string, str):
                escaped_string = escaped_string.replace(char, "\\" + char)
            else:
                escaped_string = tf.strings.regex_replace(
                    escaped_string, "\\" + char, "\\" + char
                )
        return escaped_string

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the StringContainsList layer.
        Used for saving and loading from a model.

        Specifically adds the string_constant_list and negation parameters to the
        config dictionary.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "string_constant_list": self.string_constant_list,
                "negation": self.negation,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType

from kamae.spark.params import SingleInputSingleOutputParams, UnixTimestampParams
from kamae.spark.transformers.base import BaseTransformer
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import CurrentUnixTimestampLayer


class CurrentUnixTimestampTransformer(
    BaseTransformer, SingleInputSingleOutputParams, UnixTimestampParams
):
    """
    Returns the current unix timestamp in either seconds or milliseconds.

    NOTE: Parity between this and its TensorFlow counterpart is very difficult at the
    millisecond level. TensorFlow provides much more precision of the timestamp,
    and has floating 64-bit precision of the unix timestamp in seconds.
    Whereas Spark 3.4.0 only supports millisecond precision (3 decimal places of unix
    timestamp in seconds). Therefore, parity is not guaranteed at this precision.

    It is recommended not to rely on parity at the millisecond level.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        unit: str = "s",
    ) -> None:
        """
        Initialises the CurrentUnixTimestamp layer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param unit: Unit of the output timestamp. Can be either "s" (or "seconds")
        for seconds or "ms" (or "milliseconds") for milliseconds. Defaults to "s".
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(unit="s")
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return None

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Returns a column of the current unix timestamp. If an array column is provided,
        we return an array column of identical structure with elements populated by
        the current unix timestamp.

        :param dataset: Input dataframe.
        :returns: Transformed dataframe.
        """
        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )

        def current_unix_timestamp() -> Column:
            """
            Returns the current unix timestamp in either seconds or milliseconds.

            :returns: Column of the current unix timestamp.
            """
            # TODO: For PySpark 3.5+ we can use unix_millis. For now, we use
            #  unix_timestamp that returns seconds (truncated so no milliseconds).
            #  In order to get milliseconds, we get the milliseconds from the current
            #  timestamp and add it to the truncated seconds.
            current_ts = F.current_timestamp()
            unix_timestamp_in_trucated_seconds = F.unix_timestamp(current_ts)
            milliseconds_str = F.date_format(current_ts, "SSS")
            milliseconds_float = milliseconds_str.cast("float") / 1000.0
            unix_timestamp_in_seconds = (
                unix_timestamp_in_trucated_seconds + milliseconds_float
            )
            return (
                unix_timestamp_in_seconds
                if self.getUnit() == "s"
                else unix_timestamp_in_seconds * 1000.0
            )

        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: current_unix_timestamp(),
        )

        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer.

        :returns: CurrentUnixTimestampLayer Tensorflow layer.
        """
        return CurrentUnixTimestampLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            unit=self.getUnit(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class SubStringDelimAtIndexLayer(BaseLayer):
    """
    Layer which splits a string tensor by a delimiter and
    returns the substring at the specified index. If the delimiter is the empty
    string, the string is split into bytes/characters.
    If the index is negative, start counting from the end of the string.
    If the index is out of bounds, the default value is returned.
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        delimiter: str = "_",
        index: int = 0,
        default_value: str = "",
        **kwargs,
    ) -> None:
        """
        Initialise the SubStringDelimAtIndexLayer layer.

        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param delimiter: String to split on. Defaults to `"_"`.
        :param index: Index of the substring to return. Defaults to `0`.
        If the index is negative, start counting from the end of the string.
        :param default_value: Value to return if index is out of bounds.
        Defaults to `""`.
        Defaults to `""`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.delimiter = delimiter
        self.index = index
        self.default_value = default_value

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @staticmethod
    def resolve_negative_indices(
        ragged_tensor: tf.RaggedTensor, index: int
    ) -> tf.Tensor:
        """
        Resolves negative indices to positive indices.

        :param ragged_tensor: Ragged tensor
        :param index: The index to resolve.
        :returns: The resolved index.
        """
        if index >= 0:
            raise ValueError("Index should be negative to resolve. Got positive index.")
        ragged_row_lengths = ragged_tensor.row_lengths(axis=-1)
        # Positive index is the length of the row + index. So that index = -1
        # resolves to the last dimension
        return tf.math.add(ragged_row_lengths, index)

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Splits the input string tensor by the delimiter and returns the substring
        at the specified index. If the index is out of bounds, the default value
        is returned.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if an iterable of tensors is passed
        in.

        :param inputs: Input tensor.
        :returns: Tensor with the substring at the specified index.
        """
        input_shape = tf.shape(inputs)
        # If the delimiter is empty, we split on bytes/characters.
        # Otherwise, we use the standard string split.
        ragged_strings_split = (
            tf.strings.split(inputs, sep=self.delimiter)
            if self.delimiter != ""
            else tf.strings.bytes_split(inputs)
        )

        if self.index >= 0:
            # The index is fully qualified, therefore, add the index + 1 to the shape
            # and then pad the ragged tensor to that shape. If the index is
            # out of bounds, it returns the default value
            index_shape = tf.constant([self.index + 1])
            input_shape = tf.concat([input_shape, index_shape], axis=0)
            return ragged_strings_split.to_tensor(
                default_value=self.default_value, shape=input_shape
            )[..., self.index]
        else:
            # The index is negative, so we need to resolve the positive index from it.
            resolved_index_tensor = self.resolve_negative_indices(
                ragged_tensor=ragged_strings_split, index=self.index
            )
            if isinstance(resolved_index_tensor, tf.RaggedTensor):
                # The resolved indices can be ragged or a regular tensor, however
                # are always rectangular since we only have a single ragged dimension,
                # and we have found the required index within this.
                resolved_index_tensor = resolved_index_tensor.to_tensor(
                    shape=tf.shape(inputs)
                )

            # Pad the ragged tensor to the maximum row_length of the ragged tensor
            # This could be different for each batch, however we return a single index
            # from it, and thus we will have consistent output shapes per batch.
            max_ragged_dim = tf.cast(
                tf.reduce_max(ragged_strings_split.row_lengths(axis=-1)), dtype=tf.int32
            )
            input_shape = tf.concat(
                [input_shape, tf.expand_dims(max_ragged_dim, axis=0)], axis=0
            )
            padded_tensor = ragged_strings_split.to_tensor(
                default_value=self.default_value, shape=input_shape
            )
            # Expand the indices to match the shape of the input
            expanded_indices = tf.expand_dims(resolved_index_tensor, axis=-1)
            # Replace negative indices with zeros temporarily, we will send these to the
            # default value as they are out of bounds
            non_negative_expanded_indices = tf.where(
                expanded_indices < 0,
                tf.constant(0, dtype=expanded_indices.dtype),
                expanded_indices,
            )
            # Gather the resolved indices from the padded tensor, send any negative
            # indices to the default value
            gathered_tensor = tf.where(
                expanded_indices >= 0,
                tf.gather(padded_tensor, non_negative_expanded_indices, batch_dims=-1),
                tf.constant(self.default_value),
            )
            # Squeeze out the extra dimension
            return tf.squeeze(gathered_tensor, axis=-1)

    def get_config(self) -> Dict[str, Any]:
        """
        Returns the config of the SubStringDelimAtIndex layer.
        Used for saving and loading from a model.

        Specifically adds the `delimiter`, `index` and `default_value` to the config.

        :returns: Dictionary of the config of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "delimiter": self.delimiter,
                "index": self.index,
                "default_value": self.default_value,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import tensorflow as tf
from sklearn.preprocessing import StandardScaler

from kamae.sklearn.params import SingleInputSingleOutputMixin
from kamae.sklearn.transformers import BaseTransformerMixin
from kamae.tensorflow.layers import StandardScaleLayer


class StandardScaleEstimator(
    StandardScaler,
    BaseTransformerMixin,
    SingleInputSingleOutputMixin,
):
    """
    Standard Scikit-Learn Estimator for use in Scikit-Learn pipelines.
    Wrapper over the existing implementation of the StandardScaler in Scikit-Learn,
    however operates on array columns and returns array columns. This is to align
    with the Spark implementation of the StandardScaler.

    Standardize features by removing the mean and scaling to unit variance.

    The standard score of a sample `x` is calculated as:

        z = (x - u) / s

    where `u` is the mean of the training samples
    and `s` is the standard deviation of the training samples
    """

    def __init__(self, input_col: str, output_col: str, layer_name: str) -> None:
        """
        Intializes a StandardScale estimator.

        :param input_col: Input column name.
        :param output_col: Output column name.
        :param layer_name: Name of the layer. Used as the name of the tensorflow layer
        """
        super().__init__(with_mean=True, with_std=True)
        self.input_col = input_col
        self.output_col = output_col
        self.layer_name = layer_name

    def fit(self, X: pd.DataFrame, y=None, **kwargs) -> "StandardScaleEstimator":
        """
        Fits the transformer to the data. Since the scikit-learn StandardScaler
        takes scalar values, we need to convert the numpy array to a list of scalars.
        This is to mimic the behavior of the Spark StandardScaler.

        In this, the input to our transformer is an array, and the output is a scaled
        array.

        :param X: Pandas dataframe to fit the transformer to.
        :param y: Not used, present here for API consistency by convention.
        :returns: Fit pipeline.
        """
        # Get array column as a list of scalars
        feature_array = X[self.input_col].tolist()
        super().fit(X=feature_array, y=y, sample_weight=None)
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Transforms the data using the transformer. Standardises the array `input_col`,
        creating a new standardised `output_col`.

        :param X: Pandas dataframe to transform.
        :param y: Not used, present here for API consistency by convention.
        :returns: Transformed data.
        """
        # Get array column as a list of scalars
        feature_array = X[self.input_col].tolist()
        # Transform the list of scalars
        transformed_list_of_scalars = super().transform(feature_array)
        # Set the output column to an array of the transformed list of scalars
        X[self.output_col] = list(transformed_list_of_scalars)
        return X

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the standard scaler transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
         that performs the standardization.
        """
        return StandardScaleLayer(
            name=self.layer_name, mean=self.mean_, variance=self.var_
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import tensorflow as tf

from kamae.tensorflow.typing import Tensor


def map_fn_w_axis(
    elems: Tensor,
    fn: Callable[[Tensor], Tensor],
    fn_output_signature: tf.dtypes.DType,
    axis: int = -1,
    parallel_iterations=None,
    swap_memory=False,
    infer_shape=True,
    name=None,
) -> Tensor:
    """
    Applies a function to a specific axis of a tensor using map_fn.
    Specifically uses `tf.transpose` and `tf.reshape` to rearrange the tensor so that
    the specified axis is preserved, the tensor is 2D and thus can be used with map_fn.

    After applying map_fn, the tensor is reshaped and transposed back to the original
    shape.

    :param elems: The input tensor.
    :param fn: The function to apply to the tensor. Must take a single tensor as input
    and return a tensor.
    :param fn_output_signature: The output signature of the function.
    :param axis: The axis to apply the function to. Defaults to -1.
    :param parallel_iterations: The number of iterations to run in parallel. Defaults to
    None.
    :param swap_memory: Whether to use memory swapping. Defaults to False.
    :param infer_shape: Whether to infer the shape of the output. Defaults to True.
    :param name: The name of the operation. Defaults to None.
    """
    # Permutation tensor that does nothing/identity
    identity_perm = tf.range(start=0, limit=tf.rank(elems))
    # Mod the axis param by the rank of the tensor and add 1. To resolve the positive
    # axis value when axis is negative.
    # Create the shift axis. We will roll the identity permutation by this amount to
    # transpose the input
    shift_axis = tf.math.mod(axis, tf.rank(elems)) + 1
    # Roll by negative shift axis. For example if
    # axis=0, shift_axis=1, identity_perm=[0, 1, 2]
    # Then transpose_perm = [1, 2, 0]

    # Transpose and reshape
    transpose_perm = tf.roll(identity_perm, shift=-shift_axis, axis=0)
    transposed_input = tf.transpose(elems, perm=transpose_perm)
    reshaped_input = tf.reshape(transposed_input, tf.stack([-1, tf.shape(elems)[axis]]))

    # Apply map_fn
    output = tf.map_fn(
        fn=fn,
        elems=reshaped_input,
        parallel_iterations=parallel_iterations,
        swap_memory=swap_memory,
        infer_shape=infer_shape,
        name=name,
        fn_output_signature=fn_output_signature,
    )

    # Undo reshape and transpose
    undo_reshaped_output = tf.reshape(output, tf.shape(transposed_input))
    undo_transpose_perm = tf.roll(identity_perm, shift=shift_axis, axis=0)
    return tf.transpose(undo_reshaped_output, perm=undo_transpose_perm)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional, Union

import pyspark.sql.functions as F
from pyspark import keyword_only
from pyspark.sql.types import ArrayType, DataType, DoubleType, FloatType

from kamae.spark.params import (
    ImputeMethodParams,
    MaskValueParams,
    SingleInputSingleOutputParams,
)
from kamae.spark.transformers import ImputeTransformer
from kamae.spark.utils import flatten_nested_arrays

from .base import BaseEstimator


class ImputeEstimator(
    BaseEstimator,
    SingleInputSingleOutputParams,
    MaskValueParams,
    ImputeMethodParams,
):
    """
    Imputation estimator for use in Spark pipelines.
    This estimator is used to calculate the chosen statistic of the input
    feature column. When fit is called it returns a ImputeTransformer
    which can be used to impute either the mean or median of a column.
    Rows are not included in the calculation of the statistic when they are
    either null or equal to the supplied mask value.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        maskValue: Optional[Union[float, int, str]] = None,
        imputeMethod: Optional[str] = None,
    ) -> None:
        """
        Initializes a ImputeEstimator estimator.
        Sets all parameters to given inputs.

        :param inputCol: Input column name to standardize.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
         in the keras model. If not set, we use the uid of the Spark transformer.
        :param maskValue: Value which to ignore, in addition to nulls, when
        computing imputation statistic.
        This is also the value that is imputed over in TF at inference.
        :param imputeMethod: Method by which to compute the value to be imputed.
        Valid values are "mean" or "median".
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(imputeMethod="mean")
        self.valid_impute_methods = ["mean", "median"]
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [FloatType(), DoubleType()]

    def _fit(self, dataset) -> "ImputeTransformer":
        """
        Fits the ImputeEstimator estimator to the given dataset.
        Calculates the imputation statistic of the input feature column and
        returns an ImputeTransformer with the statistic set.

        :param dataset: Pyspark dataframe to fit the estimator to.
        :returns: ImputeTransformer instance with impute value set.
        """
        imputeMethod = self.getImputeMethod()

        if imputeMethod == "mean":
            estimator_fn = F.mean
        elif imputeMethod == "median":
            estimator_fn = F.median

        input_column_type = self.get_column_datatype(dataset, self.getInputCol())
        input_col_an_array = isinstance(input_column_type, ArrayType)

        # If the column input is an array then we need to flatten and explode it to
        # calculate the impute value
        if input_col_an_array:
            # Flatten the array to a single array
            flattened_array_col = flatten_nested_arrays(
                column=F.col(self.getInputCol()), column_data_type=input_column_type
            )
            processed_input_col = F.explode(flattened_array_col).alias(
                self.uid + "_input_col"
            )
        else:
            processed_input_col = F.col(self.getInputCol()).alias(
                self.uid + "_input_col"
            )

        imputeValue = (
            dataset.select(processed_input_col)
            .select(
                F.when(
                    (F.col(self.uid + "_input_col") == F.lit(self.getMaskValue()))
                    | (F.col(self.uid + "_input_col").isNull()),
                    None,
                )
                .otherwise(F.col(self.uid + "_input_col"))
                .alias("input_col")
            )
            .agg(estimator_fn("input_col"))
            .collect()[0][0]
        )

        return ImputeTransformer(
            inputCol=self.getInputCol(),
            outputCol=self.getOutputCol(),
            layerName=self.getLayerName(),
            inputDtype=self.getInputDtype(),
            outputDtype=self.getOutputDtype(),
            imputeValue=imputeValue,
            maskValue=self.getMaskValue(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import ArrayType, DataType

from kamae.spark.params import AutoBroadcastParams, MultiInputSingleOutputParams
from kamae.spark.utils import (
    broadcast_scalar_column_to_array_with_inner_singleton_array,
    get_array_nesting_level,
    nested_arrays_zip,
    nested_transform,
)
from kamae.tensorflow.layers import ArrayConcatenateLayer

from .base import BaseTransformer


class ArrayConcatenateTransformer(
    BaseTransformer,
    MultiInputSingleOutputParams,
    AutoBroadcastParams,
):
    """
    ArrayConcatenate Spark Transformer for use in Spark pipelines.
    This transformer assembles multiple columns into a single array column.
    """

    @keyword_only
    def __init__(
        self,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        autoBroadcast: Optional[bool] = False,
    ) -> None:
        """
        Initialize a ArrayConcatenateTransformer transformer.

        :param inputCols: List of input column names.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param autoBroadcast: If True, the Keras transformer will broadcast scalar
        inputs to the biggest rank. Default is False.
        WARNING: This modifies only the Keras layer, not the Spark transformer!
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(autoBroadcast=False)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return None

    @staticmethod
    def array_concatenate_transform(
        input_cols: List[Column],
        input_col_datatypes: List[DataType],
    ) -> Column:
        """
        Similar to the transform helpers in the utils module,but for the
        ArrayConcatenateTransformer. This function is specific to this transformer
        because whilst it still broadcasts scalars to arrays, we do not want to repeat
        the innermost array elements. Thus, each array can have different innermost size
        for concatenation. Other implementations perform elementwise computations across
        nested arrays and so require that all arrays are exactly the same shape.

        Caters for the case, where the input columns are:

        1. All scalars.
        2. All (possibly nested) arrays.
        3. A mix of scalars and (possibly nested) arrays.

        If all inputs are scalars, applies concat(array()) directly to the input
        columns. If inputs are a mixture of scalars and arrays, broadcasts any scalars
        to the size of the arrays in all N-1 dimensions, and creates an array of size 1
        in the final Nth dimension. Then zips the arrays into a single nested array
        column, stopping before the innermost (Nth) array dimension. Finally, concats
        along this Nth dimension.

        Example:

        Input 1: 1.0 (scalar)
        Input 2: [
                    [[2.0, 3.0], [4.0, 5.0]],
                    [[6.0, 7.0], [8.0, 9.0]]
                ] (nested 3D array)
        Input 3: 10.0 (scalar)
        Input 4: [
                    [[11.0, 12.0, 13.0], [14.0, 15.0, 16.0]],
                    [[17.0, 18.0, 19.0], [20.0, 21.0, 22.0]]
                ] (nested 3D array)

        Input 1 & 3 are broadcast to the size of the arrays up to N-1th dim, giving:
        Input 1: [
                    [[1.0], [1.0]],
                    [[1.0], [1.0]]
                ]
        Input 3: [
                    [[10.0], [10.0]],
                    [[10.0], [10.0]]
                ]
        Then zipped together to give:
        [
            [
                {
                    "input1": [1.0],
                    "input2": [2.0, 3.0],
                    "input3": [10.0],
                    "input4": [11.0, 12.0, 13.0]
                },
                {
                    "input1": [1.0],
                    "input2": [4.0, 5.0],
                    "input3": [10.0],
                    "input4": [14.0, 15.0, 16.0]
                }
            ],
            [
                {
                    "input1": [1.0],
                    "input2": [6.0, 7.0],
                    "input3": [10.0],
                    "input4": [17.0, 18.0, 19.0]
                },
                {
                    "input1": [1.0],
                    "input2": [8.0, 9.0],
                    "input3": [10.0],
                    "input4": [20.0, 21.0, 22.0]
                }
            ]
        ]
        Then concats along the Nth dimension to give final output:
        [
            [
                [1.0, 2.0, 3.0, 10.0, 11.0, 12.0, 13.0],
                [1.0, 4.0, 5.0, 10.0, 14.0, 15.0, 16.0]
            ],
            [
                [1.0, 6.0, 7.0, 10.0, 17.0, 18.0, 19.0],
                [1.0, 8.0, 9.0, 10.0, 20.0, 21.0, 22.0]
            ]
        ]

        :param input_cols: List of input columns.
        :param input_col_datatypes: List of input column datatypes.
        :returns: Output column.
        """
        input_col_names = [f"input{i}" for i in range(len(input_cols))]
        if all(
            [not isinstance(datatype, ArrayType) for datatype in input_col_datatypes]
        ):
            # All inputs are scalars.
            # Apply the concat, wrapping each scalar in an array.
            return F.concat(*[F.array(x) for x in input_cols])

        if all([isinstance(datatype, ArrayType) for datatype in input_col_datatypes]):
            # All inputs are arrays. Zip the arrays together into a single column.
            nesting_level = get_array_nesting_level(column_dtype=input_col_datatypes[0])
            zipped_array_column = nested_arrays_zip(
                columns=input_cols,
                nest_level=nesting_level - 1,
                column_names=input_col_names,
            )
        else:
            # Inputs are a mix of scalars and arrays.
            # Broadcast the scalars to the size of the arrays.
            scalar_columns = [
                (idx, col_w_datatype[0])
                for idx, col_w_datatype in enumerate(
                    zip(input_cols, input_col_datatypes)
                )
                if not isinstance(col_w_datatype[1], ArrayType)
            ]
            array_columns_w_types = [
                (idx, col_w_datatype[0], col_w_datatype[1])
                for idx, col_w_datatype in enumerate(
                    zip(input_cols, input_col_datatypes)
                )
                if isinstance(col_w_datatype[1], ArrayType)
            ]
            # Broadcast the scalar to the size of the arrays. Use the first array column
            # to determine the size of the broadcasted scalar. Assumes all arrays are
            # of the same size in the N-1 dimensions
            broadcasted_scalars = [
                (
                    idx,
                    broadcast_scalar_column_to_array_with_inner_singleton_array(
                        scalar_column=scalar_column,
                        array_column=array_columns_w_types[0][1],
                        array_column_datatype=array_columns_w_types[0][2],
                    ),
                )
                for idx, scalar_column in scalar_columns
            ]
            # Resort the array columns and the broadcasted scalars,
            # so they match the order of the input columns.
            columns_w_idx = [
                (idx, column) for idx, column, _ in array_columns_w_types
            ] + broadcasted_scalars
            sorted_columns = [
                column for idx, column in sorted(columns_w_idx, key=lambda x: x[0])
            ]
            nesting_level = get_array_nesting_level(
                column_dtype=array_columns_w_types[0][2]
            )
            zipped_array_column = nested_arrays_zip(
                columns=sorted_columns,
                nest_level=nesting_level - 1,
                column_names=input_col_names,
            )

        # Create the nested transform function that applies the function to the zipped
        # array column.
        nested_func = nested_transform(
            func=lambda x: F.concat(*[x[c] for c in input_col_names]),
            nest_level=nesting_level - 1,
        )
        return nested_func(zipped_array_column)

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transform the input dataset. Creates a new column named outputCol which is a
        concatenated array of all input columns.

        :param dataset: Pyspark DataFrame to transform.
        :returns: Transformed pyspark dataFrame.
        """
        input_col_names = self.getInputCols()
        input_cols = [F.col(c) for c in input_col_names]
        input_col_datatypes = [
            self.get_column_datatype(dataset=dataset, column_name=c)
            for c in input_col_names
        ]

        output_col = self.array_concatenate_transform(
            input_cols=input_cols,
            input_col_datatypes=input_col_datatypes,
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that concatneates the input tensors.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that concatenates the input tensors.
        """
        return ArrayConcatenateLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            axis=-1,
            auto_broadcast=self.getAutoBroadcast(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import MultiInputSingleOutputParams
from kamae.spark.utils import multi_input_single_output_scalar_transform
from kamae.tensorflow.layers import StringConcatenateLayer

from .base import BaseTransformer


class StringConcatenateParams(Params):
    """
    Mixin class containing separator parameter needed for string concatenation
    transforms.
    """

    separator = Param(
        Params._dummy(),
        "separator",
        "Value to use as a separator when joining the strings.",
        typeConverter=TypeConverters.toString,
    )

    def setSeparator(self, value: str) -> "StringConcatenateParams":
        """
        Sets the separator parameter.

        :param value: String value to use as a separator when joining the strings.
        :returns: Instance of class mixed in.
        """
        return self._set(separator=value)

    def getSeparator(self) -> str:
        """
        Gets the separator parameter.

        :returns: String value to use as a separator when joining the strings.
        """
        return self.getOrDefault(self.separator)


class StringConcatenateTransformer(
    BaseTransformer,
    MultiInputSingleOutputParams,
    StringConcatenateParams,
):
    """
    String Concatenate Spark Transformer for use in Spark pipelines.
    This transformer takes in multiple columns and concatenates them together into a
    single column using a separator. Input columns must be of type string.
    """

    @keyword_only
    def __init__(
        self,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        separator: str = "_",
    ) -> None:
        """
        Initializes the string concatenate transformer.
        :param inputCols: columns to concatenate together. Must be of type string.
        :param outputCol: column to output the concatenated string to.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param separator: String to use as a separator when joining the strings.
        If not provided, underscore `_` is used.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self._setDefault(separator="_")
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        where the value is the result of concatenating the values of the input columns.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        separator = self.getSeparator()

        input_col_names = self.getInputCols()
        input_cols = [F.col(c) for c in input_col_names]

        input_datatypes = [
            self.get_column_datatype(dataset=dataset, column_name=input_col_name)
            for input_col_name in input_col_names
        ]

        output_col = multi_input_single_output_scalar_transform(
            input_cols=input_cols,
            input_col_datatypes=input_datatypes,
            input_col_names=input_col_names,
            func=lambda x: F.concat_ws(
                separator, *[x[input_col_name] for input_col_name in input_col_names]
            ),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the concatenate transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs a concatenation.
        """
        return StringConcatenateLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            separator=self.getSeparator(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union

import tensorflow as tf
from sklearn.base import BaseEstimator, TransformerMixin

from kamae.sklearn.params import InputOutputExtractor, LayerNameMixin


class BaseTransformerMixin(ABC, LayerNameMixin, InputOutputExtractor):
    """
    Mixin abstract class defining methods needed for all kamae scikit-learn
    transformers.
    """

    def __init__(self, **kwargs) -> None:
        """
        Initializes the transformer.
        """
        super().__init__()

    @abstractmethod
    def get_tf_layer(self) -> Union[tf.keras.layers.Layer, List[tf.keras.layers.Layer]]:
        """
        Gets the tensorflow layer to be used in the model.
        This is the only abstract method that must be implemented.
        :returns: Tensorflow Layer
        """
        raise NotImplementedError

    def construct_layer_info(self) -> Dict[str, Any]:
        """
        Constructs the layer info dictionary.
        Contains the layer name, the tensorflow layer, and the inputs and outputs.
        This is used when constructing the pipeline graph.

        :returns: Dictionary containing layer information such as
        name, tensorflow layer, inputs, and outputs.
        """
        inputs, outputs = self.get_layer_inputs_outputs()
        return {
            "name": self.layer_name,
            "layer": self.get_tf_layer(),
            "inputs": inputs,
            "outputs": outputs,
        }


class BaseTransformer(BaseTransformerMixin, BaseEstimator, TransformerMixin, ABC):
    """
    Abstract class for all scikit-learn transformers. Specifically, this class extends
    the required scikit-learn classes BaseEstimator and TransformerMixin adding in the
    kamae BaseTransformerMixin which defines the methods needed to work with the kamae
    pipeline graph.

    The reason we keep this separate from the BaseTransformerMixin (which is not done
    for Spark) is because on the scikit-learn side we want to allow the ability to
    inherit from existing scikit-learn classes (such as the StandardScaler). In these
    cases the existing class already inherits from BaseEstimator and TransformerMixin
    and so only needs the BaseTransformerMixin (to add kamae specific functionality).
    If you try and inherit these classes twice (once from the existing scikit-learn
    class and once from BaseTransformer) you will get an error. Therefore, we keep
    these separate.

    If you are building an entirely new transformer, then you can inherit from this
    class directly, to save you from having to inherit from BaseEstimator and
    TransformerMixin.

    In Spark, all existing (core) implementations are built in Scala and ported to
    Python. In this case, the ability to re-use existing Spark transformers is very
    difficult and not worth the effort. You can see that for the StandardScaleEstimator
    the logic does not depend on the existing Spark StandardScaler.

    Therefore, we have a single BaseTransformer class for use by all Spark
    transformers.
    """



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.ml.param.shared import (
    HasInputCol,
    HasInputCols,
    HasOutputCol,
    HasOutputCols,
)
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable

from kamae.utils import DType

from .name import HasLayerName
from .utils import InputOutputExtractor


class HasInputDtype(Params):
    """
    Mixin class for a transformer input datatype.
    """

    inputDtype = Param(
        Params._dummy(),
        "inputDtype",
        "Input datatype of the transformer",
        typeConverter=TypeConverters.toString,
    )

    def setInputDtype(self, value: str) -> "HasInputDtype":
        """
        Sets the parameter inputDtype to the given string value.

        :param value: String to set the inputDtype parameter to.
        :raises ValueError: If the input dtype is not supported.
        :returns: Instance of class mixed in.
        """
        dtype_enums = [dtype.dtype_name for dtype in DType]
        if value not in dtype_enums:
            raise ValueError(
                f"""Input dtype {value} not supported.
                Supported dtypes are: {dtype_enums}"""
            )
        return self._set(inputDtype=value)

    def getInputDtype(self) -> str:
        """
        Gets the value of the inputDtype parameter.
        :returns: Input datatype.
        """
        return self.getOrDefault(self.inputDtype)

    def getInputTFDtype(self) -> Optional[str]:
        """
        Gets the tensorflow datatype string from the inputDtype parameter.
        Uses the DType enum within Kamae to map the inputDtype to the tensorflow
        datatype string.
        :returns: String of the tensorflow datatype.
        """
        input_dtype = self.getInputDtype()
        if input_dtype is None:
            return None
        dtypes_map = {dtype.dtype_name: dtype.tf_dtype.name for dtype in DType}
        return dtypes_map[input_dtype]


class HasOutputDtype(Params):
    """
    Mixin class for a transformer output datatype.
    """

    outputDtype = Param(
        Params._dummy(),
        "outputDtype",
        "Output datatype of the transformer",
        typeConverter=TypeConverters.toString,
    )

    def setOutputDtype(self, value: str) -> "HasOutputDtype":
        """
        Sets the parameter outputDtype to the given string value.

        :param value: String to set the outputDtype parameter to.
        :raises ValueError: If the output dtype is not supported.
        :returns: Instance of class mixed in.
        """
        dtype_enums = [dtype.dtype_name for dtype in DType]
        if value not in dtype_enums:
            raise ValueError(
                f"""Output dtype {value} not supported.
                Supported dtypes are: {dtype_enums}"""
            )
        return self._set(outputDtype=value)

    def getOutputDtype(self) -> str:
        """
        Gets the value of the outputDtype parameter.
        :returns: Output datatype.
        """
        return self.getOrDefault(self.outputDtype)

    def getOutputTFDtype(self) -> Optional[str]:
        """
        Gets the tensorflow datatype string from the outputDtype parameter.
        Uses the DType enum within Kamae to map the outputDtype to the tensorflow
        datatype string.
        :returns: String of the tensorflow datatype.
        """

        output_dtype = self.getOutputDtype()
        if output_dtype is None:
            return None
        dtypes_map = {dtype.dtype_name: dtype.tf_dtype.name for dtype in DType}
        return dtypes_map[output_dtype]


class SingleInputParams(HasInputCol):
    """
    Mixin class containing set methods for the single input column scenario.
    """

    def setInputCol(self, value: str) -> "SingleInputParams":
        """
        Sets the parameter inputCol to the given string value.

        :param value: String to set the inputCol parameter to.
        :returns: Instance of class mixed in.
        """
        return self._set(inputCol=value)


class MultiInputParams(HasInputCols):
    """
    Mixin class containing set methods for the multiple input columns scenario.
    """

    def setInputCols(self, value: List[str]) -> "MultiInputParams":
        """
        Sets the parameter inputCols to the given list of strings.

        :param value: List of strings to set the inputCols parameter to.
        :returns: Instance of class mixed in.
        """
        return self._set(inputCols=value)


class SingleOutputParams(HasLayerName, HasOutputCol, HasOutputDtype):
    """
    Mixin class containing set methods for the single output column scenario.
    """

    def setLayerName(self, value: str) -> "SingleOutputParams":
        """
        Sets the parameter layerName to the given string value.
        Throws an error if the value is the same as the output column name,
        as this causes issues when constructing the pipeline graph.

        :param value: String to set the layerName parameter to.
        :returns: Instance of class mixed in.
        """
        if self.hasParam("outputCol") and self.isDefined("outputCol"):
            if value == self.getOutputCol():
                raise ValueError("Layer name and output column name must be different.")
        return self._set(layerName=value)

    def setOutputCol(self, value: str) -> "SingleOutputParams":
        """
        Sets the parameter outputCol to the given string value.
        Throws an error if the value is the same as the layer name,
        as this causes issues when constructing the pipeline graph.

        :param value: String to set the outputCol parameter to.
        :returns: Instance of class mixed in.
        """
        if self.hasParam("layerName") and self.isDefined("layerName"):
            if value == self.getLayerName():
                raise ValueError("Layer name and output column name must be different.")
        return self._set(outputCol=value)


class MultiOutputParams(HasLayerName, HasOutputCols, HasOutputDtype):
    """
    Mixin class containing set methods for the multiple output columns scenario.
    """

    def setLayerName(self, value: str) -> "MultiOutputParams":
        """
        Sets the parameter layerName to the given string value.
        Throws an error if the value is the same as one of the output column names,
        as this causes issues when constructing the pipeline graph.

        :param value: String to set the layerName parameter to.
        :returns: Instance of class mixed in.
        """
        if self.hasParam("outputCol") and self.isDefined("outputCols"):
            if value in self.getOutputCols():
                raise ValueError(
                    "Layer name and output column names must be different."
                )
        return self._set(layerName=value)

    def setOutputCols(self, value: List[str]) -> "MultiOutputParams":
        """
        Sets the parameter outputCols to the given list of strings.
        Throws an error if one of the output column names is the same as the layer name,
        as this causes issues when constructing the pipeline graph.

        :param value: List of strings to set the outputCols parameter to.
        :returns: Instance of class mixed in.
        """
        if self.hasParam("layerName") and self.isDefined("layerName"):
            if self.getLayerName() in value:
                raise ValueError(
                    "Layer name and output column names must be different."
                )
        return self._set(outputCols=value)


class SingleInputSingleOutputParams(
    SingleInputParams,
    SingleOutputParams,
    InputOutputExtractor,
    DefaultParamsReadable,
    DefaultParamsWritable,
):
    """
    Mixin class containing set methods for the single input
    and single output column scenario.
    """


class SingleInputMultiOutputParams(
    SingleInputParams,
    MultiOutputParams,
    InputOutputExtractor,
    DefaultParamsReadable,
    DefaultParamsWritable,
):
    """
    Mixin class containing set methods for the single input
    and multiple output columns scenario.
    """


class MultiInputSingleOutputParams(
    MultiInputParams,
    SingleOutputParams,
    InputOutputExtractor,
    DefaultParamsReadable,
    DefaultParamsWritable,
):
    """
    Mixin class containing set methods for the multiple input
    and single output column scenario.
    """


class MultiInputMultiOutputParams(
    MultiInputParams,
    MultiOutputParams,
    InputOutputExtractor,
    DefaultParamsReadable,
    DefaultParamsWritable,
):
    """
    Mixin class containing set methods for the multiple input
    and multiple output columns scenario.
    """



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Type, Union, cast

import keras_tuner as kt
import tensorflow as tf
from pyspark.ml import PipelineModel
from pyspark.ml.pipeline import (
    PipelineModelReader,
    PipelineModelWriter,
    PipelineSharedReadWrite,
)
from pyspark.ml.util import DefaultParamsReader, MLWriter

from kamae.graph import PipelineGraph
from kamae.spark.transformers import BaseTransformer

if TYPE_CHECKING:
    from pyspark.ml._typing import ParamMap


class KamaeSparkPipelineModel(PipelineModel):
    """
    KamaeSparkPipelineModel is a subclass of pyspark.ml.PipelineModel that is used to
    chain together LayerTransformers. It maintains the same functionality
    as pyspark.ml.PipelineModel e.g. serialisation.
    """

    def __init__(self, stages: List[BaseTransformer]):
        """
        Initialises the KamaeSparkPipelineModel object.

        :param stages: List of LayerTransformers to chain together.
        """
        super().__init__(stages=stages)
        self.stages = stages

    def copy(self, extra: Optional["ParamMap"] = None) -> "KamaeSparkPipelineModel":
        """
        Creates a copy of the KamaeSparkPipelineModel object.

        :param extra: Additional optional params to copy to new pipeline model.
        :returns: KamaeSparkPipelineModel object.
        """
        if extra is None:
            extra = dict()
        stages = [stage.copy(extra) for stage in self.stages]
        return KamaeSparkPipelineModel(stages)

    def write(self) -> MLWriter:
        """
        Uses the KamaeSparkPipelineModelWriter class to write the pipeline model to a
        persistent storage path.

        :returns: KamaeSparkPipelineModelWriter object.
        """
        return KamaeSparkPipelineModelWriter(self)

    @classmethod
    def read(cls) -> "KamaeSparkPipelineModelReader":
        """
        Uses the KamaeSparkPipelineModelReader class to read a pipeline model from a
        persistent storage path.

        :returns: KamaeSparkPipelineModelReader object.
        """
        return KamaeSparkPipelineModelReader(cls)

    def get_all_tf_layers(self) -> List[tf.keras.layers.Layer]:
        """
        Gets a list of all tensorflow layers in the pipeline model.

        :returns: List of tensorflow layers within the pipeline model.
        """
        return [stage.get_tf_layer() for stage in self.stages]

    def expand_pipeline_stages(self) -> List[BaseTransformer]:
        """
        Expands the pipeline stages to include all nested pipeline stages.
        If the pipeline stage is itself a pipeline model, it will be expanded
        recursively.

        :returns: List of all pipeline stages flattened to transformer level.
        """
        expanded_stages = []
        for stage in self.stages:
            if isinstance(stage, KamaeSparkPipelineModel):
                # Recursively expand the pipeline stages.
                expanded_stages.extend(stage.expand_pipeline_stages())
            else:
                expanded_stages.append(stage)
        return expanded_stages

    def build_keras_model(
        self,
        tf_input_schema: Union[List[tf.TypeSpec], List[Dict[str, Any]]],
        output_names: Optional[List[str]] = None,
    ) -> tf.keras.Model:
        """
        Builds a keras model from the pipeline model using the PipelineGraph
        helper class.

        :param tf_input_schema: List of dictionaries containing the input schema for
        the model. Specifically the name, shape and dtype of each input.
        These will be passed as is to the Keras Input layer.
        :param output_names: Optional list of output names for the Keras model. If
        provided, only the outputs specified are used as model outputs.
        :returns: Keras model.
        """
        stage_dict = {
            stage.getOrDefault("layerName"): stage.construct_layer_info()
            for stage in self.expand_pipeline_stages()
        }
        pipeline_graph = PipelineGraph(stage_dict=stage_dict)
        return pipeline_graph.build_keras_model(
            tf_input_schema=tf_input_schema, output_names=output_names
        )

    def get_keras_tuner_model_builder(
        self,
        tf_input_schema: Union[List[tf.TypeSpec], List[Dict[str, Any]]],
        hp_dict: Dict[str, List[Dict[str, Any]]],
        output_names: Optional[List[str]] = None,
    ) -> Callable[[kt.HyperParameters], tf.keras.Model]:
        """
        Builds a keras tuner model builder (function) from the pipeline model
        using the PipelineGraph helper class.

        :param tf_input_schema: List of dictionaries containing the input schema for
        the model. Specifically the name, shape and dtype of each input.
        These will be passed as is to the Keras Input layer.
        :param hp_dict: Dictionary containing the hyperparameters for the model.
        :param output_names: Optional list of output names for the Keras model. If
        provided, only the outputs specified are used as model outputs.
        :returns: Keras tuner model builder (function).
        """
        stage_dict = {
            stage.getOrDefault("layerName"): stage.construct_layer_info()
            for stage in self.stages
        }
        pipeline_graph = PipelineGraph(stage_dict=stage_dict)
        return pipeline_graph.get_keras_tuner_model_builder(
            tf_input_schema=tf_input_schema, hp_dict=hp_dict, output_names=output_names
        )


class KamaeSparkPipelineModelReader(PipelineModelReader):
    """
    Util class for reading a pipeline model from a persistent storage path.
    """

    def __init__(self, cls: Type["KamaeSparkPipelineModel"]):
        super().__init__(cls=cls)

    def load(self, path: str) -> "KamaeSparkPipelineModel":
        """
        Loads a pipeline model from a given path.

        :param path: Path to stored pipeline model.
        :returns: KamaeSparkPipelineModel object.
        """
        metadata = DefaultParamsReader.loadMetadata(path, self.sc)
        uid, stages = PipelineSharedReadWrite.load(metadata, self.sc, path)
        return KamaeSparkPipelineModel(
            stages=cast(List[BaseTransformer], stages)
        )._resetUid(uid)


class KamaeSparkPipelineModelWriter(PipelineModelWriter):
    """
    Util class for writing a pipeline model to a persistent storage path.
    """

    def __init__(self, instance: "KamaeSparkPipelineModel"):
        super().__init__(instance=instance)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import StringMapLayer

from .base import BaseTransformer


class StringMapTransformerParams(Params):
    """
    Mixin class containing StringMatchValues and StringReplaceValues
    needed for string map layers.
    """

    stringMatchValues = Param(
        Params._dummy(),
        "stringMatchValues",
        "String match constant to use in string replace.",
        typeConverter=TypeConverters.toListString,
    )

    stringReplaceValues = Param(
        Params._dummy(),
        "stringReplaceValues",
        "String replace constant to use in string replace.",
        typeConverter=TypeConverters.toListString,
    )

    defaultReplaceValue = Param(
        Params._dummy(),
        "defaultReplaceValue",
        """
        Default value to replace the unmatched strings with.
        If None, the original string is kept unchanged.
        """,
        typeConverter=TypeConverters.toString,
    )

    def setStringMatchValues(self, value: List[str]) -> "StringMapTransformerParams":
        """
        Sets the stringMatchValues parameter.

        :param value: List of string match constants.
        :returns: Instance of class mixed in.
        """
        if value is None or len(value) == 0:
            raise ValueError("stringMatchValues cannot be empty.")
        return self._set(stringMatchValues=value)

    def getStringMatchValues(self) -> List[str]:
        """
        Gets the stringMatchValues parameter.

        :returns: List of string match constants.
        """
        return self.getOrDefault(self.stringMatchValues)

    def setStringReplaceValues(self, value: List[str]) -> "StringMapTransformerParams":
        """
        Sets the stringReplaceValues parameter.

        :param value: List of string replace constants.
        :returns: Instance of class mixed in.
        """
        if value is None or len(value) == 0:
            raise ValueError("stringReplaceValues cannot be empty.")
        return self._set(stringReplaceValues=value)

    def getStringReplaceValues(self) -> List[str]:
        """
        Gets the stringReplaceValues parameter.

        :returns: List of string replace constants.
        """
        return self.getOrDefault(self.stringReplaceValues)

    def setDefaultReplaceValue(self, value: str) -> "StringMapTransformerParams":
        """
        Sets the defaultReplaceValue parameter.

        :param value: Default value to replace the unmatched strings with.
        If None, the original string is kept unchanged.
        :returns: Instance of class mixed in.
        """
        return self._set(defaultReplaceValue=value)

    def getDefaultReplaceValue(self) -> str:
        """
        Gets the defaultReplaceValue parameter.

        :returns: Default value to replace the unmatched strings with.
        """
        return self.getOrDefault(self.defaultReplaceValue)


class StringMapTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    StringMapTransformerParams,
):
    """
    String Map Spark Transformer for use in Spark Pipelines.
    This transformer replaces a list of strings with the respective mapping value.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        stringMatchValues: Optional[List[str]] = None,
        stringReplaceValues: Optional[List[str]] = None,
        defaultReplaceValue: Optional[str] = None,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initializes an StringMapTransformer transformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param stringMatchValues: List of string match constants.
        :param stringReplaceValues: List of string replace constants.
        :param defaultReplaceValue: Default value to replace the unmatched strings with.
        If None, the original string is kept unchanged.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(
            stringMatchValues=None,
            stringReplaceValues=None,
            defaultReplaceValue=None,
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _string_map(self, column: Column) -> Column:
        """
        Helper function to create a string map expression.

        :param column: Column to apply the string map operation to.
        :returns: Column with string map operation applied.
        """
        col_expr: Column = None
        for match_value, replace_value in zip(
            self.getStringMatchValues(), self.getStringReplaceValues()
        ):
            if col_expr is None:
                col_expr = F.when(column == F.lit(match_value), replace_value)
            else:
                col_expr = col_expr.when(column == F.lit(match_value), replace_value)
        if self.getDefaultReplaceValue() is not None:
            col_expr = col_expr.otherwise(self.getDefaultReplaceValue())
        else:
            col_expr = col_expr.otherwise(column)
        return col_expr

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which contains the result of the string map operation.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        if self.getStringMatchValues() is None or self.getStringReplaceValues() is None:
            raise ValueError(
                "stringMatchValues and stringReplaceValues cannot be None."
            )
        if len(self.getStringMatchValues()) != len(self.getStringReplaceValues()):
            raise ValueError(
                "Length of stringMatchValues and stringReplaceValues must be equal."
            )
        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: self._string_map(x),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the StringMapLayer transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
        performs a string replace operation.
        """
        return StringMapLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            string_match_values=self.getStringMatchValues(),
            string_replace_values=self.getStringReplaceValues(),
            default_replace_value=self.getDefaultReplaceValue(),
        )
